#include "B4RDefines.h"
 
namespace B4R {

	//returns a B4R byte array containing date/time representing the seconds given
	ArrayByte* B4RSoftRtc::Sec2DT(ULong sec)  {
		ArrayByte* datetime = CreateStackMemoryObject(ArrayByte);
		datetime->data = StackMemory::buffer + StackMemory::cp;
		datetime->length = 8;
		B4R::StackMemory::cp += 8;
		
		byte dt[8];
		SecAsDateTime(sec,dt);
		for (byte i = 0;i < 8;i++) {
			((byte*)datetime->data)[i] = dt[i];
		}
		
		return datetime;
	}
	
	//private function
	//convert seconds into an 8 byte array containing date/time
	void B4RSoftRtc::SecAsDateTime(ULong sec, byte* dt)  {
		ULong days = sec / 86400ul;   //NB: suffix ul forces constant to unsigned-long
		
		ULong dow = ((days+6)%7)+1;    //day of week  1=monday, 7=sunday (never 0)

		sec = sec - (days*86400ul);
		ULong hrs = sec / 3600ul;  //?? was int
		sec = sec - (hrs*3600ul);
		ULong min = sec / 60ul;  //?? was int
		sec = sec - (min*60ul);

		ULong lpy = days / 1461ul;
		days = days - (lpy*1461ul);
		
		ULong ya = 0;
		ULong y = 0;
		if (days>=366)  {
			ya=1;  //the first year is a leap year
			days = days - 366;
			y = (days/365ul);  //remainder years
			days = days - (y*365ul);
		}
		y = y + ya;

		ULong m;

		days++;		//day zero becomes the 1st	
		if (days<=31)  {m=1;} else { 
		  days = days - 31;
		  if (days<=28)  {m=2;} else {
			 if ((y==0) && (days == 29))  {m=2;} else {
				if (y==0)  {
					days = days - 29;	
				} else {
					days = days - 28;
				} 
			    if (days<=31)  {m=3;} else {
				  days=days-31; 
				  if (days<=30)  {m=4;} else {
				    days=days-30;
				    if (days<=31)	 {m=5;} else {
					  days=days-31;
					  if (days<=30)  {m=6;} else {
					    days=days-30; 
						if (days<=31)  {m=7;} else {
						  days=days-31;
						  if (days<=31)  {m=8;} else {
						    days=days-31;
							if (days<=30)  {m=9;} else {
							  days=days-30;
							  if (days<=31)  {m=10;} else {
								days=days-31;
								if (days<=30)  {m=11;} else {
								  days=days-30;
								   m=12;
								}
							  }
							}
						  }
						}
					 } 
				  }
				} 
			  }
		    }
		  }  
		}
		
		dt[0] = 21;  //century
		dt[1] = y + (lpy*4) + 12;  //add 12 years back in
		dt[2] = m;
		dt[3] = days;
		dt[4] = hrs;
		dt[5] = min;
		dt[6] = sec;
		dt[7] = dow;  //day of the week
	}

	//returns the current date/time as a B4R byte array
	bool B4RSoftRtc::ReadClock(ArrayByte* datetime)  {
		if (SetAt_millis==0)  {
			return false;  //clock has not been set
		}

		ULong sec;
		sec = (millis() - SetAt_millis)/1000;
		sec = sec + Set_DateTime_AsSeconds;

		byte dt[8];
		SecAsDateTime(sec,dt);
		for (byte i = 0;i < 8;i++) {
			((byte*)datetime->data)[i] = dt[i];
		}

		return true;  //success
	}
	
	//returns a seconds for the B4R byte array containing a date/time
	ULong B4RSoftRtc::DT2Sec(ArrayByte* datetime)  {
		ULong sec;
		ULong days;
		ULong m;
		ULong y;
		
		int dt[8];
		for (byte i = 0;i < 8;i++) {
			 dt[i] = ((byte*)datetime->data)[i];
		}
		
		sec = dt[6] + dt[5]*60 + dt[4]*3600ul;  //convert time to seconds
		
		days = dt[3]-1;  //1st becomes day zero
		m = dt[2];
		y = dt[1] - 12;  //subtract 12 years (to reduce the seconds count)
		
		ULong lpy = y/4;
		days = days + (lpy*1461ul);
		y = y - (lpy*4);
		
		if (y>0)  {
			days = days + 366;
			y--;
		} else {
			if (m>2) days++;
		}
		
		days = days + (y*365ul);

		days = days + (m-1)*30;	
		if (m>1) days++;
		if (m>2) days=days-2;
		if (m>3) days++;
		if (m>5) days++;
		if (m>7) days++;
		if (m>8) days++;
		if (m>10) days++;
	
		sec = sec + (days*86400ul);
		
		return sec;
	}
	
	void B4RSoftRtc::SetClock(ArrayByte* datetime, 
							int offset_hours, int offset_minutes)  {

		ULong sec = DT2Sec(datetime);
		ULong ofsec = ((offset_hours*60)+offset_minutes)*60;
		sec = sec + ofsec;
		
		Set_DateTime_AsSeconds = sec;
		SetAt_millis = millis();
		
		if (tick_every_second)  {
			start_one_second_tick();
		}
	}

	static union FunctionUnion fncu;
	void B4RSoftRtc::Initialize(void (*Event) (byte idx), bool tick_every_sec) {
		fncu.TimerFunction = tick;
		this->event = Event;
	
		for (byte i = 0;i < 10;i++) {
			this->enabled[i] = 0;  //0=not active
		}
		
		tick_every_second = tick_every_sec;
	}

	void B4RSoftRtc::start_one_second_tick() {
		counter++;
		if (counter==0)  counter=1;
		this->enabled[0] = counter;
		scheduler.add(millis() + 1000, fncu, counter, this);
	}
	
	bool B4RSoftRtc::SetAlarm(byte Index, ArrayByte* datetime) {
		if ((Index==0) || (Index>9))  {
			return false;   //index out of range (1 to 9 only)
		}
		
		ULong CurrentAsSec;
		CurrentAsSec = (millis() - SetAt_millis)/1000;
		CurrentAsSec = CurrentAsSec + Set_DateTime_AsSeconds;
		ULong AlmAsSec = DT2Sec(datetime);
		if (AlmAsSec<=CurrentAsSec)  {
			return false;   //date/time has already passed
		}
		ULong Interval = (AlmAsSec - CurrentAsSec);
		Interval = Interval * 1000;
		
		counter++;
		if (counter==0)  counter=1;
		this->enabled[Index] = counter;
		scheduler.add(millis() + Interval, fncu, counter, this);
		return true;  //success
	}
	bool B4RSoftRtc::CancelAlarm(byte Index)  {
		if ((Index==0) || (Index>9))  {
			return false;   //index out of range (0 to 9 only)
		}
		this->enabled[Index] = 0;  //0=not active
		return true;  //success
	}
	
	//called on alarm event
	void B4RSoftRtc::tick(byte tag, void* target) {
		B4RSoftRtc* me = (B4RSoftRtc*)target;

		for (byte i = 0;i < 10;i++) {
			if (me->enabled[i] == tag)  {
				me->enabled[i] = 0;  //0=not active
				if (i==0)  {
					me->start_one_second_tick();  //and the next second, and so on
				}
				me->event(i);  //raise B4R alarm event with it's "Index" (0 to 9)
			}
		}
	}

	B4RString* B4RSoftRtc::ShortDate(ArrayByte* datetime) {
		PrintToMemory pm;
		B4RString* s = B4RString::PrintableToString(NULL);
		
		byte v;
		v =((byte*)datetime->data)[3];  //day
		if (v<10) pm.print(0);
		pm.print(v);
		
		StackMemory::buffer[StackMemory::cp++] = (byte)'/';		
		
		v =((byte*)datetime->data)[2];  //month
		v=((v-1)*3);
		String months = "JanFebMarAprMayJunJulAugSepOctNovDec";
		for (byte i = 0;i < 3;i++) {
			StackMemory::buffer[StackMemory::cp++] = (byte)months[v+i];		
		}
		
		StackMemory::buffer[StackMemory::cp++] = (byte)'/';	
		
		v =((byte*)datetime->data)[1];  //year
		pm.print(v);
		StackMemory::buffer[StackMemory::cp++] = 0;  //end of string
		
		return s;  //short date as string
	}

		B4RString* B4RSoftRtc::Time24(ArrayByte* datetime) {
		PrintToMemory pm;
		B4RString* s = B4RString::PrintableToString(NULL);
		
		byte v;
		v =((byte*)datetime->data)[4];  //hours
		if (v<10) pm.print(0);
		pm.print(v);
		
		StackMemory::buffer[StackMemory::cp++] = (byte)':';		
		
		v =((byte*)datetime->data)[5];  //minutes
		if (v<10) pm.print(0);
		pm.print(v);
		
		StackMemory::buffer[StackMemory::cp++] = (byte)':';		
		
		v =((byte*)datetime->data)[6];  //seconds
		if (v<10) pm.print(0);
		pm.print(v);

		StackMemory::buffer[StackMemory::cp++] = 0;  //end of string
		
		return s;  //24hr time
	}

		B4RString* B4RSoftRtc::Time12(ArrayByte* datetime) {
		PrintToMemory pm;
		B4RString* s = B4RString::PrintableToString(NULL);
		
		byte h;
		h =((byte*)datetime->data)[4];  //hours
		byte v=h%12;
		if (v==0) v=12;
		if (v<10) pm.print(0);
		pm.print(v);

		StackMemory::buffer[StackMemory::cp++] = (byte)':';		
		
		v =((byte*)datetime->data)[5];  //minutes
		if (v<10) pm.print(0);
		pm.print(v);
		
		if (h<12) {
			StackMemory::buffer[StackMemory::cp++] = (byte)'a';			
		} else {
			StackMemory::buffer[StackMemory::cp++] = (byte)'p';			
		}
		StackMemory::buffer[StackMemory::cp++] = (byte)'m';		
		
		StackMemory::buffer[StackMemory::cp++] = 0;  //end of string
		
		return s;  //12hr time
	}

	B4RString* B4RSoftRtc::ShortDOW(ArrayByte* datetime) {
		//day of week
		PrintToMemory pm;
		B4RString* s = B4RString::PrintableToString(NULL);
		
		byte v;
		v =((byte*)datetime->data)[7];  //day of week
		v=(v*3);
		String dow = "   MonTueWedThuFriSatSun";
		for (byte i = 0;i < 3;i++) {
			StackMemory::buffer[StackMemory::cp++] = (byte)dow[v+i];		
		}
		StackMemory::buffer[StackMemory::cp++] = 0;  //end of string
		
		return s;  //short day-of-week as string
	}
	
}
